from enum import unique, Enum


@unique
class Ex(Enum):
    """
    The 'Ex' enum provides a variety of values that make it easy to extract specific data from a html-tag.
    An 'Ex' enum value can be used to set 'extract' parameter of a 'BlueMoss' instance.

    Note that setting the 'extract' parameter of a 'BlueMoss' instance only has an effect when the BlueMoss instance
    is a leaf node, i.e. its 'nodes' parameter is an empty list.

    :param TEXT: Yields the text contained in a html-tag excluding the text contained in child html-tags.
    :param FULL_TEXT: Yields the text contained in a html-tag including the text contained in child html-tags.
    :param BS4_TAG: Yields the current html tag as a BeautifulSoup object.
    :param TAG_AS_STRING: Yields the current html tag as a string.
    :param LXML_HTML_ELEMENT: Yields the current tag as a lxml.html.HtmlElement object.
    :param HREF: Yields the href attribute of a tag.
    :param HREF_QUERY:
        Yields the query-string part of a href attribute of a tag.
        E.g. the tag <a href="https://example.com?foo=bar">https://example.com</a> will yield "foo=bar".
    :param HREF_DOMAIN:
        Yields the entire domain including subdomains of a href attribute of a tag.
        E.g. the tag <a href="https://sub1.sub2.example.com?foo=bar">https://sub1.sub2.example.com</a>
        will yield "sub1.sub2.example.com".
    :param HREF_ENDPOINT:
        Yields the endpoint without the query-string of a href attribute of a tag.
        E.g. the tag <a href="https://example.com/products/page1?foo=bar">https://example.com</a>
        will yield "/products/page1".
    :param HREF_ENDPOINT_WITH_QUERY:
        Yields the endpoint with the query-string of a href attribute of a tag.
        E.g. the tag <a href="https://example.com/products/page1?foo=bar">https://example.com</a>
        will yield "/products/page1?foo=bar".
    :param HREF_BASE_DOMAIN:
        Yields the base-domain of a href attribute of a tag.
        E.g. the tag <a href="https://sub1.sub2.example.com?foo=bar">https://sub1.sub2.example.com</a>
        will yield "example.com". If no base-domain can be found in the href attribute it will yield the None value.
    :param HREF_QUERY_PARAMS:
        Yields the query-params as a dict of the href attribute of a tag.
        E.g. the tag <a href="https://example.com/products/page1?foo1=bar1&foo2=bar2&foo1=bar3">https://example.com</a>
        will yield the dictionary {"foo1": ["bar1", "bar3"], "foo2": ["bar2"]}.
    """

    # text
    TEXT = 'text'
    FULL_TEXT = 'full_text'

    # tag
    BS4_TAG = 'tag'
    TAG_AS_STRING = 'tag_as_string'
    LXML_HTML_ELEMENT = 'lxml_html_etree_element'

    # href
    HREF = 'href'
    HREF_QUERY = 'href_query'
    HREF_DOMAIN = 'href_domain'
    HREF_ENDPOINT = 'href_endpoint'
    HREF_BASE_DOMAIN = 'href_base_domain'
    HREF_QUERY_PARAMS = 'href_query_params'
    HREF_ENDPOINT_WITH_QUERY = 'href_endpoint_with_query'


__all__ = ['Ex']



from __future__ import annotations
from . import utils
from lxml.html import HtmlElement
from typing import Type, Any, cast
from lxml import html as lxml_html
from .classes import Ex, BlueMoss, Range, JsonifyWithTag


def extract(moss: BlueMoss, html: str) -> Any:
    """Main function of bluemoss: Scrapes :param html with the scraping-recipe defined by :param moss.

    :param moss: BlueMoss object. Defines how to scrape :param html.
    :param html: HTML doc to be scraped.
    :rtype: Any
    :return: data extracted from :param html
    """
    # noinspection PyTypeChecker
    tag: HtmlElement = lxml_html.fromstring(html)
    return _extract(0, moss, tag)


def _extract(level: int, moss: BlueMoss, tag: HtmlElement) -> Any:
    """Internal main function, which is recursively called in order to scrape :param tag using :param moss.

    :param moss: Defines how to scrape :param tag.
    :param tag: The html-tag to be scraped.
    :param level:
        The depth level of the current Bluemoss node :param moss (0 at root-level).
        The reason for why we keep track of the depth level of @param moss in relation to it's root node, is because
        we want to allow for the root node (BlueMoss object, level = 0) to define a value for its parameter 'key'
        in which case we want to return a dict with just one key-value-pair (key pointing to the extracted data).
    :rtype: Any
    :return: data extracted from :param html
    """

    if moss.no_xpath:
        # if no xpath was defined, we continue to work with the current tag
        matched_tags = [tag]
    else:
        # find matching tags
        tags: list[HtmlElement | str] = tag.xpath(moss.full_xpath)
        # filter the matched tags against moss.filter
        matched_tags = _filter_matched_tags(moss, tags)

    if len(matched_tags) == 0:
        # We either did not find any matching tags or we found matching tags but did not filter any of them.
        return moss.transform(None if moss.find_single_tag else [])

    if moss.target or moss.nodes:
        val = [_build_target(level, moss, tag) for tag in matched_tags]
    else:
        # len(moss.nodes) == 0 and moss.target is None
        val = [_extract_from_leaf_node(moss, tag) for tag in matched_tags]

    if isinstance(moss.filter, int):
        # moss.filter is an int => we want to filter for a single tag
        assert len(val) == 1
        val = val[0]  # type: ignore

    if level == 0 and moss.key is not None:
        # About this if-condition:
        #     1) level == 0: @param moss is most outer parent node (root node)
        #     2) moss.key is not None: extracted data (val) shall be wrapped in a dictionary (since level == 0)

        # Why do we only want to get here if @param level is 0?
        #     - If level > 0, then moss.key will be used in the function '_build_target'.
        #     - So, a way to make use of moss.key when level == 0, is to build a dictionary
        #       where moss.key points to @param val.
        return {moss.key: moss.transform(val)}

    return moss.transform(val)


def _filter_matched_tags(
    moss: BlueMoss, tags: list[HtmlElement | str]
) -> list[HtmlElement | str | None]:
    """Filter the matched tags."""

    if isinstance(moss.filter, int):
        # moss.filter is an int => we want to filter for a single tag
        try:
            return [tags[moss.filter]]
        except IndexError:
            return []

    # moss.filter is not an int => we want to filter for multiple tags

    if isinstance(moss.filter, Range):
        # moss.filter is a Range object => filter for multiple subsequent tags
        return moss.filter.filter(tags)

    if isinstance(moss.filter, list):
        # moss.filter is a list of integers => filter for the tags whose index is contained in moss.filter
        _tags: list[HtmlElement | str | None] = []
        for idx in moss.filter:
            try:
                _tags.append(tags[idx])
            except IndexError:
                _tags.append(None)
        return _tags

    # moss.filter is None => return all matched tags
    return tags


def _build_target(
    level: int, moss: BlueMoss, tag: HtmlElement | str | None
) -> list[Any] | dict[str, Any] | Type[Any] | None:
    """
    Builds the target instance (list, dict or class/dataclass instance) of :param moss.
    Called by '_extract' function in case :param moss defines its 'nodes' parameter.

    :rtype: list | dict | Class (class/dataclass instance)
    :return:
        1. Class instance if the 'target' param of :param moss is not None.
        2. List or Dict if the 'target' param of :param moss is None:
            2.1: Dict if all nodes in :param moss define their 'key' param.
            2.2: List if all nodes have their 'key' param set to None.
    """

    if tag is None:
        return None

    if moss.target is None:
        if moss.keys_in_nodes:
            # the target is a dict
            return {
                node.key: _extract(level + 1, node, tag)  # type: ignore
                for node in moss.nodes
            }
        # the target is a list
        return [_extract(level + 1, node, tag) for node in moss.nodes]

    # the target is a class/dataclass

    # :param values: dictionary to instantiate moss.target
    values: dict[str, Any] = {
        node.key: _extract(level + 1, node, tag)  # type: ignore
        for node in moss.nodes
    }

    if issubclass(moss.target, JsonifyWithTag):
        # If moss.target inherits from JsonifyWithTag, the intention is to provide
        # the current html-tag @param tag to our target-instance using the key '_tag'.
        values |= {'_tag': tag}

    # noinspection PyArgumentList
    return moss.target(**values)  # type: ignore


def _extract_from_leaf_node(
    moss: BlueMoss, tag: HtmlElement | str | None
) -> Any:
    """
    Extracts data from a leaf node.
    A leaf node is a BlueMoss object with an empty 'nodes' list (moss.nodes == []).

    :rtype: Any
    :return: data extracted from :param moss.
    """

    if tag is None:
        return None

    if isinstance(tag, str):
        # @param tag is a string if moss.xpath endswith with e.g. /@href,  /@class or /text(),
        # i.e. when we do not select for a tag but for a tag-property, like text or an attribute.
        return str(tag)

    if isinstance(moss.extract, str):
        # extract the value of the tag-attribute defined by moss.extract
        return tag.get(moss.extract)

    # extract data from @param tag by match-casing moss.extract
    if moss.extract == Ex.FULL_TEXT:
        return utils.clean_text(tag.text_content().strip())
    elif moss.extract == Ex.LXML_HTML_ELEMENT:
        return tag
    elif moss.extract == Ex.TEXT:
        return tag.text.strip()
    elif moss.extract == Ex.BS4_TAG:
        return utils.lxml_etree_to_bs4(tag)
    elif moss.extract == Ex.TAG_AS_STRING:
        return cast(Any, utils.lxml_etree_to_bs4(tag)).prettify()
    elif moss.extract == Ex.HREF:
        return tag.get('href')
    elif moss.extract == Ex.HREF_QUERY:
        return utils.get_url_query(tag.get('href'))
    elif moss.extract == Ex.HREF_DOMAIN:
        return utils.get_domain(tag.get('href'))
    elif moss.extract == Ex.HREF_ENDPOINT:
        return utils.get_endpoint(tag.get('href'))
    elif moss.extract == Ex.HREF_BASE_DOMAIN:
        return utils.get_base_domain(tag.get('href'))
    elif moss.extract == Ex.HREF_QUERY_PARAMS:
        return utils.get_url_query_params(tag.get('href'))
    elif moss.extract == Ex.HREF_ENDPOINT_WITH_QUERY:
        return utils.get_endpoint_with_query(tag.get('href'))
    else:
        raise NotImplementedError


__all__ = ['extract']



from __future__ import annotations
from enum import Enum
from json import dumps
from bs4 import BeautifulSoup
from lxml.html import HtmlElement
from dataclasses import dataclass
from collections import OrderedDict
from datetime import date, datetime
from collections import OrderedDict
from .constants import FOUR_DIVS_HTML as HTML
from src.bluemoss.utils.html import lxml_etree_to_bs4
from src.bluemoss import Jsonify, JsonifyWithTag, Root, Ex, extract


class Job(Enum):
    ENGINEER = 'Engineer'
    FARMER = 'Farmer'
    LAWYER = 'Lawyer'


@dataclass
class Address(Jsonify):
    city: str
    street: str
    zip_code: str
    country_code: str


@dataclass
class Skill(Jsonify):
    name: str
    started_at: date
    experience_level: int


@dataclass
class Employer:
    name: str
    website: str
    headquarters: OrderedDict[str, str]


@dataclass
class Person(Jsonify):
    name: str
    email: str
    birthday: date
    address: Address
    employer: Employer
    last_contacted: datetime
    job: Job
    skills: list[Skill]
    languages: set[str]
    _last_updated: datetime

    @property
    def estimated_age(self) -> int:
        today: date = date.today()
        return today.year - self.birthday.year

    @property
    def dict(self) -> dict:
        return super().dict | {'age': self.estimated_age}


def test_person_dict():
    address = Address(
        city='Los Angeles',
        street='789 Creative Blvd',
        zip_code='90001',
        country_code='US',
    )

    skill1 = Skill(
        name='Software Development',
        started_at=date(2010, 6, 1),
        experience_level=5,
    )

    skill2 = Skill(
        name='Database Management',
        started_at=date(2012, 8, 10),
        experience_level=4,
    )

    person = Person(
        name='Charlie Davis',
        email='charlie.davis@example.com',
        birthday=date(1988, 10, 20),
        address=address,
        last_contacted=datetime(2023, 10, 21, 19, 24, 35),
        job=Job.ENGINEER,
        skills=[skill1, skill2],
        languages={'English', 'French'},
        employer=Employer(
            name='Company',
            website='https://company.com',
            headquarters=OrderedDict([('USA', 'New York'), ('UK', 'London')]),
        ),
        _last_updated=datetime.now(),
    )

    expected_dict: OrderedDict = OrderedDict(
        [
            ('name', 'Charlie Davis'),
            ('email', 'charlie.davis@example.com'),
            ('birthday', '1988-10-20'),
            (
                'address',
                OrderedDict(
                    [
                        ('city', 'Los Angeles'),
                        ('street', '789 Creative Blvd'),
                        ('zip_code', '90001'),
                        ('country_code', 'US'),
                    ]
                ),
            ),
            (
                'employer',
                OrderedDict(
                    [
                        ('name', 'Company'),
                        ('website', 'https://company.com'),
                        (
                            'headquarters',
                            OrderedDict(
                                [('USA', 'New York'), ('UK', 'London')]
                            ),
                        ),
                    ]
                ),
            ),
            ('last_contacted', '2023-10-21 19:24:35'),
            ('job', 'Engineer'),
            (
                'skills',
                [
                    OrderedDict(
                        [
                            ('name', 'Software Development'),
                            ('started_at', '2010-06-01'),
                            ('experience_level', 5),
                        ]
                    ),
                    OrderedDict(
                        [
                            ('name', 'Database Management'),
                            ('started_at', '2012-08-10'),
                            ('experience_level', 4),
                        ]
                    ),
                ],
            ),
            ('languages', ['English', 'French']),
            ('age', 35),
        ]
    )

    assert person.dict == expected_dict
    assert str(person) == dumps(person.dict, indent=4)


def test_jsonify_with_tag():
    @dataclass
    class Tag(JsonifyWithTag):
        pass

    tag_instance: Tag = extract(Root('div', target=Tag), HTML)
    assert isinstance(tag_instance, Tag)
    assert isinstance(tag_instance.lxml_tag, HtmlElement)
    assert tag_instance.source_line == 4

    bs4_tag: BeautifulSoup = extract(Root('div', extract=Ex.BS4_TAG), HTML)
    assert isinstance(bs4_tag, BeautifulSoup)

    assert tag_instance.bs4_tag == bs4_tag
    assert lxml_etree_to_bs4(tag_instance.lxml_tag) == bs4_tag



from __future__ import annotations
import inspect
from typing import Type, Any
from dataclasses import fields, MISSING, is_dataclass


def get_required_class_init_params(cls: Type[Any] | None) -> set[str]:
    """
    :param cls: class reference
    :rtype: set[str]
    :return: The set of required initialization parameters for the given class.
    """
    if cls is None:
        return set()
    return get_all_class_init_params(cls) - get_optional_class_init_params(cls)


def get_all_class_init_params(cls: Type[Any] | None) -> set[str]:
    """
    :param cls: class reference
    :rtype: set[str]
    :return: The set of all initialization parameters for the given class.
    """
    if cls is None:
        return set()
    return set(inspect.getfullargspec(cls.__init__).args[1:])


def get_optional_class_init_params(cls: Type[Any] | None) -> set[str]:
    """
    :param cls: class reference
    :rtype: set[str]
    :return: The set of optional initialization parameters for the given class.
    """
    if cls is None:
        return set()
    if is_dataclass(cls):
        return {
            f.name
            for f in fields(cls)
            if f.default != MISSING or f.default_factory != MISSING
        }
    return {
        name
        for name, param in inspect.signature(cls.__init__).parameters.items()
        if param.default != inspect.Parameter.empty
    }


__all__ = [
    'get_all_class_init_params',
    'get_optional_class_init_params',
    'get_required_class_init_params',
]



from __future__ import annotations
import pytest
from .constants import HIGH_NESTING_LEVEL_HTML as HTML
from src.bluemoss import Root, Range, Ex, InvalidXpathException, extract


def test_valid_paths():
    Root()
    Root('')
    Root('a')
    Root('//a')


def test_invalid_path_1():
    with pytest.raises(InvalidXpathException):
        Root('body///section')


def test_invalid_path_2():
    with pytest.raises(InvalidXpathException):
        Root('//')


def test_invalid_path_3():
    with pytest.raises(InvalidXpathException):
        Root('///a')


def test_non_existent_path():
    xpath: str = 'h4'
    assert extract(Root(xpath), HTML) is None
    assert extract(Root(xpath, filter=3), HTML) is None
    assert extract(Root(xpath, filter=None), HTML) == []


def test_nested_path():
    xpath: str = 'body//section//li/a/../h3'
    assert extract(Root(xpath, filter=None), HTML) == [
        'Service 1',
        'Service 2',
    ]
    assert extract(Root(xpath), HTML) == 'Service 1'
    assert extract(Root(xpath, filter=0), HTML) == 'Service 1'
    assert extract(Root(xpath, filter=1), HTML) == 'Service 2'
    assert extract(Root(xpath, filter=Range(0)), HTML) == [
        'Service 1',
        'Service 2',
    ]
    assert extract(Root(xpath, filter=Range(1)), HTML) == ['Service 2']
    assert extract(Root(xpath, filter=Range(4)), HTML) == []


def test_nested_path_with_tag_attribute_search():
    xpath: str = "body//section//li/a[@href='/info']/../h3"
    assert extract(Root(xpath, filter=None), HTML) == ['Service 2']
    assert extract(Root(xpath), HTML) == 'Service 2'


def test_nested_path_with_text_search():
    xpath: str = "html//section/ul//h3[text()='Service 1']/../a"
    assert extract(Root(xpath), HTML) == 'Learn More'
    assert (
        extract(Root(xpath, extract=Ex.HREF_ENDPOINT), HTML) == '/learn-more'
    )



from __future__ import annotations
from bs4 import BeautifulSoup
from src.bluemoss import utils
from datetime import date, datetime
from dataclasses import dataclass, field
from lxml import html as lxml_html, etree


@dataclass
class PersonDataClass:
    name: str
    birthday: date
    address: str | None = field(default=None)
    friends: list[PersonDataClass] = field(default_factory=list)
    last_updated: datetime | None = field(default=None, init=False)


class PersonClass:
    def __init__(
        self,
        name: str,
        birthday: date,
        friends: list[PersonClass],
        address: str | None = None,
    ):
        self.name = name
        self.address = address
        self.friends = friends
        self.birthday = birthday
        self.last_updated = datetime.now()


def test_person_class_instantiation():
    PersonClass(
        name='test_name',
        birthday=date.today(),
        address='test_address',
        friends=[
            PersonClass(name='test_name_2', birthday=date.today(), friends=[])
        ],
    )


def test_person_data_class_instantiation():
    PersonDataClass(
        name='test_name',
        birthday=date.today(),
        address='test_address',
        friends=[
            PersonDataClass(
                name='test_name_2', birthday=date.today(), friends=[]
            )
        ],
    )


# general.py
def test_get_required_init_params():
    assert utils.get_required_class_init_params(None) == set()
    assert utils.get_required_class_init_params(PersonDataClass) == {
        'name',
        'birthday',
    }
    assert utils.get_required_class_init_params(PersonClass) == {
        'name',
        'birthday',
        'friends',
    }


def test_get_all_init_params():
    assert utils.get_all_class_init_params(None) == set()
    init_params: set[str] = {'name', 'birthday', 'address', 'friends'}
    assert utils.get_all_class_init_params(PersonDataClass) == init_params
    assert utils.get_all_class_init_params(PersonClass) == init_params


def test_get_optional_init_params():
    assert utils.get_optional_class_init_params(None) == set()
    assert utils.get_optional_class_init_params(PersonDataClass) == {
        'address',
        'friends',
        'last_updated',
    }
    assert utils.get_optional_class_init_params(PersonClass) == {'address'}


def test_clean_text():
    # fmt: off
    text: str = """
        Hello world!
                        Here.

        This is a test.



            Hello world! 
 

        Test.
            Hello world!


    """
    # fmt: on
    assert (
        utils.clean_text(text)
        == 'Hello world!\nHere.\nThis is a test.\nHello world!\nTest.\nHello world!'
    )
    assert utils.clean_text(None) is None


def test_get_infix():
    text: str = '    Hello world!  '
    assert utils.get_infix(text, 'Hello', '') == ''
    assert utils.get_infix(text, 'Hello', 'Galaxy') is None
    assert utils.get_infix(text, 'He', 'orld!') == 'llo w'
    assert utils.get_infix('', 'He', 'orld!') is None
    assert (
        utils.get_infix(text, 'NonExistentPrefix', 'NonExistentSuffix') is None
    )


# html.py
def test_etree_to_bs4():
    html: str = '<html><body><p>Hello world!</p></body></html>'
    assert utils.lxml_etree_to_bs4(html) is None
    soup: BeautifulSoup = utils.lxml_etree_to_bs4(lxml_html.fromstring(html))  # type: ignore
    assert isinstance(soup, BeautifulSoup)
    assert str(soup) == html


def test_remove_tags():
    html: str = """<html><body><p>Hello world!</p><h1>Headline</h1><p>This is code.</p></body></html>"""

    for (tags_to_remove, expected_html) in [
        ('p', '<html><body><h1>Headline</h1></body></html>'),
        ('body', '<html></html>'),
        ('html', ''),
    ]:
        soup = BeautifulSoup(html, 'html.parser')
        utils.remove_tags_from_soup(soup, tags_to_remove)
        assert str(soup) == expected_html


def test_etree_to_string():
    html: str = '<html><body><p>Hello world!</p></body></html>'
    elem: etree.Element = lxml_html.fromstring(html)
    assert utils.lxml_etree_to_string(elem) == html


# url.py

URLS: list[str | None] = [
    'https://example.com/path/to/resource/',
    'https://sub1.example.net/resource',
    'https://sub1.sub2.example.org/path/to/resource2?query=val',
    'https://www.sub3.example.com?p=q',
    'https://example2.org/path/to/resource3?p=v&q=w&p=z&a=b',
    '    ',
    '',
    None,
]


def test_get_base_domain():
    base_domains: list[str] = [
        'example.com',
        'example.net',
        'example.org',
        'example.com',
        'example2.org',
        None,
        None,
        None,
    ]
    for i, url in enumerate(URLS):
        assert utils.get_base_domain(url) == base_domains[i]


def test_get_domain():
    domains: list[str] = [
        'example.com',
        'sub1.example.net',
        'sub1.sub2.example.org',
        'sub3.example.com',
        'example2.org',
        None,
        None,
        None,
    ]
    for i, url in enumerate(URLS):
        assert utils.get_domain(url) == domains[i]


def test_get_endpoint():
    endpoints: list[str] = [
        '/path/to/resource',
        '/resource',
        '/path/to/resource2',
        None,
        '/path/to/resource3',
        None,
        None,
        None,
    ]
    for i, url in enumerate(URLS):
        assert utils.get_endpoint(url) == endpoints[i]


def test_get_url_query():
    queries: list[str | None] = [
        None,
        None,
        'query=val',
        'p=q',
        'p=v&q=w&p=z&a=b',
        None,
        None,
        None,
    ]
    for i, url in enumerate(URLS):
        assert queries[i] == utils.get_url_query(url)


def test_get_endpoint_with_query():
    endpoints_with_query: list[str] = [
        '/path/to/resource',
        '/resource',
        '/path/to/resource2?query=val',
        '?p=q',
        '/path/to/resource3?p=v&q=w&p=z&a=b',
        None,
        None,
        None,
    ]
    for i, url in enumerate(URLS):
        assert endpoints_with_query[i] == utils.get_endpoint_with_query(url)


def test_get_url_query_params():
    query_params: list[dict[str, str]] = [
        {},
        {},
        {'query': ['val']},
        {'p': ['q']},
        {'p': ['v', 'z'], 'q': ['w'], 'a': ['b']},
        {},
        {},
        {},
    ]
    for i, url in enumerate(URLS):
        assert query_params[i] == utils.get_url_query_params(url)



from __future__ import annotations
import builtins
from .extract import Ex
from .range import Range
from inspect import isclass
from functools import cached_property
from typing import Callable, Type, Any
from dataclasses import dataclass, field
from ..utils import (
    is_valid_xpath,
    get_all_class_init_params,
    get_required_class_init_params,
)


@dataclass(frozen=True)
class BlueMoss:
    """
    A BlueMoss object is a recipe that describes:

    1. What data to extract from which tags.
    2. How to further transform the extracted data.
    3. Where to place the transformed data, e.g. in a class, dataclass, dict, or list.

    :param xpath:
        The xpath used to find e.g. a single tag, multiple tags, a tag-attribute value, etc.
    :param xpath_prefix:
        An optional prefix to prepend to the `xpath`, e.g. '/', '//', './' or './/'. Useful for defining subclasses
        of BlueMoss, like 'Root' and 'Node', which might use common xpath prefixes such as `.//` for `Root` and `//`
        for `Node`. This simplifies the creation of BlueMoss objects by avoiding manual prefixing of every `xpath`.
    :param target:
        Specifies the class or dataclass to wrap the extracted data. Requires all objects in `nodes` to set their
        'key' parameter to a non-None value. If `target` is None but `nodes` is populated, the extracted data will
        be wrapped in a dict or list, depending on whether all instances in `nodes` set their `key` parameter.
    :param key:
        An optional key for the extracted data. Depending on the parent BlueMoss object's settings, a `key` will
        place the data in a dict or a class/dataclass.
    :param nodes:
        A list of child nodes (BlueMoss instances) for extracting multiple data points from tags matched against
        `full_path`. All BlueMoss instances in `nodes` should either set a 'key' value or none at all.
    :param extract:
        A string or an Ex enum value denoting the data to extract from an HTML tag. If `extract` is a string,
        the function will try to extract the value of the HTML tag attribute specified by `extract`.
    :param transform:
        A Callable for further processing of the extracted data.
    :param filter:
        Enables filtering for specific tags matched against `full_path`. It can be:
        1. int: From all matched tags, the int value denotes the index of the desired HTML tag.
        2. list[int]: A list of ints representing indices of matched tags of interest.
        3. Range: A Range object (see ./range.py for documentation).
        4. None: No filtering is applied; data is extracted from all matched tags.
    """

    xpath: str = ''
    xpath_prefix: str = ''
    key: str | None = None
    target: Type[Any] | None = None
    extract: Ex | str = Ex.FULL_TEXT
    filter: int | list[int] | Range | None = 0
    transform: Callable[[Any], Any] = lambda x: x
    nodes: list[Node] = field(default_factory=list)

    @property
    def no_xpath(self) -> bool:
        """
        :rtype: bool
        :return: True if no xpath was provided, False otherwise.
        """
        return self.xpath == ''

    @property
    def find_single_tag(self) -> bool:
        """
        :rtype: bool
        :return: True if only one tag should be filtered out of the matched tags, False otherwise.
        """
        return isinstance(self.filter, int)

    @cached_property
    def full_xpath(self) -> str:
        """
        :rtype: str
        :return: The xpath to use when querying html with the extract function.
        """
        return f'{self.xpath_prefix}{self.xpath}'

    @cached_property
    def keys_in_nodes(self) -> set[str]:
        """
        If this set is not empty, then we use these keys in the dictionary that instantiates a target object:
            1) If the target value of the parent-node (if any) is not None,
               then we instantiate that class using the dict.
            2) If the target value of the parent node is None, then we simply use the dict as the target object.

        :rtype: set[str]
        :return: The set of all keys in the :param nodes.
        """
        return {c.key for c in self.nodes if c.key is not None}

    @cached_property
    def target_class_name(self) -> str | None:
        return self.target.__name__ if self.target else None

    def __post_init__(self) -> None:
        if not (self.no_xpath or is_valid_xpath(self.full_xpath)):
            """Check the provided xpath for syntactical correctness."""
            raise InvalidXpathException(self)

        if not (
            all([c.key is not None for c in self.nodes])
            or all([c.key is None for c in self.nodes])
        ):
            """
            Asserts that either all (or none) of the BlueMoss instances in :param nodes set the 'key' param.
            - In case none of the instances set a key value, the target object will be a list.
            - I case all of the instances set a value for the 'key' param , then the target is either a dict
              or a class/dataclass, depending on whether the parent node has set a value for the 'target' param.
              If it hasn't, then the target is a dict.
            """
            raise PartialKeysException(self)

        if self.target is None:
            """Early return in case no target was provided."""
            return

        if self.target.__name__ in dir(builtins):
            """Make sure the value of the 'target' param references a class, which is not a builtin class."""
            raise InvalidTargetTypeException(self)

        if not self.keys_in_nodes.issubset(
            get_all_class_init_params(self.target)
        ):
            """
            Make sure that all keys provided in the instances of the 'nodes'
            param are valid init params for the target class.
            """
            raise InvalidKeysForTargetException(self)

        if not get_required_class_init_params(self.target).issubset(
            self.keys_in_nodes
        ):
            """
            Make sure that the instances in the 'nodes' param cover all mandatory
            initialization parameters of the target class.
            """
            raise MissingTargetKeysException(self)


@dataclass(frozen=True)
class Root(BlueMoss):
    """
    BlueMoss subclass which comes with :param xpath_prefix set to '//'.
    This class is suited to be used as the most outer parent node in a BlueMoss tree.
    """

    xpath_prefix: str = field(default='//', init=False)


@dataclass(frozen=True)
class Node(BlueMoss):
    """
    BlueMoss subclass which comes with :param xpath_prefix set to './/'.
    This class is suited to be used as a child node in a BlueMoss tree as the preceding '.' char
    tells us to match :param full_path against the tag(s) matched in the parent node.
    """

    xpath_prefix: str = field(default='.//', init=False)


class InvalidXpathException(Exception):
    def __init__(self, moss: BlueMoss):
        message: str = (
            f'\n{moss.full_xpath} seems to be an invalid xpath. '
            f'\nFeel free to use ChatGPT to check if your path is compatible with the XPath 1.0 syntax.'
            f'\nNote that xpath queries using XPath syntax of any version higher than 1.0 are not supported.'
        )
        super().__init__(message)


class PartialKeysException(Exception):
    def __init__(self, moss: BlueMoss):
        message: str = (
            f"\n\nSome Node instances in your nodes list have a key attribute set while others don't."
            f'\nYou can either provide a list of nodes where ALL instances set a key, or NO instances do.'
        )
        super().__init__(message)


class InvalidTargetTypeException(Exception):
    def __init__(self, moss: BlueMoss):
        message: str = (
            f"\n\nThe type of your target '{moss.target_class_name}' is not a custom class nor a dataclass."
            f'\nMake sure the target either refers to a class or a dataclass.'
            f'\n\nPro Tips:'
            f'\n1) If you want your target to be a DICT, '
            f"then you don't have to set the 'target' parameter. All you have to do is to provide keys "
            f"with all Node instances in your 'nodes' list."
            f"\n2) If you want your target to be a LIST, then you also don't need to set your 'target' parameter. "
            f"Just provide the list of Node instances in your 'nodes' list "
            f"without any of the Node instances having set the 'key' parameter."
        )
        super().__init__(message)


class InvalidKeysForTargetException(Exception):
    def __init__(self, moss: BlueMoss):
        invalid_keys: set[str] = {
            key
            for key in moss.keys_in_nodes
            if key not in get_all_class_init_params(moss.target)
        }
        message: str = (
            f'A Node instance in your nodes list defines a key that is no valid init '
            f'parameter for your target {moss.target_class_name}: {invalid_keys.pop()}'
        )
        super().__init__(message)


class MissingTargetKeysException(Exception):
    def __init__(self, moss: BlueMoss):
        missing_keys: list[str] = sorted(
            list(
                (
                    get_required_class_init_params(moss.target)
                    - moss.keys_in_nodes
                )
            )
        )
        message: str = f"\n\nMissing keys in nodes list for target '{moss.target_class_name}': {missing_keys}"
        super().__init__(message)


__all__ = [
    'BlueMoss',
    'Root',
    'Node',
    'InvalidXpathException',
    'PartialKeysException',
    'MissingTargetKeysException',
    'InvalidTargetTypeException',
    'InvalidKeysForTargetException',
]



import pytest
from lxml import etree
from bs4 import BeautifulSoup
from src.bluemoss import Root, Ex, extract
from .constants import WITH_LINKS_HTML as HTML


def test_text_extraction():
    moss = Root('p', extract=Ex.TEXT)
    assert extract(moss, HTML) == 'Lorem 1'

    moss = Root('div', extract=Ex.TEXT)
    assert extract(moss, HTML) == 'Ipsum 2'


def test_full_text_extraction():
    moss = Root('div', extract=Ex.FULL_TEXT)
    assert extract(moss, HTML) == 'Ipsum 2\nLorem 2\nLink 2'


def test_tag_extraction():
    moss = Root('div', extract=Ex.BS4_TAG)
    tag: BeautifulSoup = extract(moss, HTML)
    assert isinstance(tag, BeautifulSoup)

    moss = Root('div')
    assert extract(moss, tag.prettify()) == 'Ipsum 2\nLorem 2\nLink 2'


def test_etree_extraction():
    # 1) extract the found tag as an etree._Element instance
    moss = Root('div', extract=Ex.LXML_HTML_ELEMENT)
    elem: etree._Element = extract(moss, HTML)
    assert isinstance(elem, etree._Element)

    # 2) transform the etree._Element into a html string and test the full-text-extraction on it
    html: str = etree.tostring(elem, method='html')
    moss = Root('div')
    assert extract(moss, html) == 'Ipsum 2\nLorem 2\nLink 2'


def test_tag_as_string_extraction():
    moss = Root('div', extract=Ex.TAG_AS_STRING)
    html: str = extract(moss, HTML)
    assert isinstance(html, str)
    moss = Root('div')
    assert extract(moss, html) == 'Ipsum 2\nLorem 2\nLink 2'


def test_href_extraction():
    moss = Root('a', filter=2, extract=Ex.HREF)
    assert extract(moss, HTML) == 'https://www.nvidia.com/link3?p=3&q=3'


def test_href_extraction_with_xpath():
    for moss in [
        Root('a/@href', filter=2),
        Root('a/@href', filter=2, extract=Ex.HREF_BASE_DOMAIN),
        Root('a/@href', filter=2, extract='some_random_tag_attribute'),
    ]:
        assert extract(moss, HTML) == 'https://www.nvidia.com/link3?p=3&q=3'


def test_href_query_extraction():
    moss = Root('a', extract=Ex.HREF_QUERY)
    assert extract(moss, HTML) == 'p=1&q=1'


def test_href_domain_extraction():
    moss = Root('a', filter=3, extract=Ex.HREF_DOMAIN)
    assert extract(moss, HTML) == 'chat.openai.com'


def test_href_base_domain_extraction():
    moss = Root('a', filter=3, extract=Ex.HREF_BASE_DOMAIN)
    assert extract(moss, HTML) == 'openai.com'


def test_href_endpoint_extraction():
    moss = Root('a', filter=3, extract=Ex.HREF_ENDPOINT)
    assert extract(moss, HTML) == '/page4/link4'


def test_href_endpoint_with_query_extraction():
    moss = Root('a', filter=3, extract=Ex.HREF_ENDPOINT_WITH_QUERY)
    assert extract(moss, HTML) == '/page4/link4?p=4&q=4&p=5'


def test_href_query_params_extraction():
    moss = Root('a', filter=3, extract=Ex.HREF_QUERY_PARAMS)
    assert extract(moss, HTML) == {'p': ['4', '5'], 'q': ['4']}


def test_extract_with_string_value():
    moss = Root('a', filter=2, extract='href')
    assert extract(moss, HTML) == 'https://www.nvidia.com/link3?p=3&q=3'


def test_extract_with_invalid_extract_value():
    moss = Root('a', extract=None)  # type: ignore
    with pytest.raises(NotImplementedError):
        extract(moss, HTML)



from __future__ import annotations
from lxml.html import HtmlElement
from dataclasses import dataclass, field


@dataclass(frozen=True)
class Range:
    """
    A Range instance can be provided to the 'filter' parameter of a 'BlueMoss' instance.
    It basically tells the 'extract' function to filter out a sequence of subsequent html-tags
    that where matched against the 'full_xpath' parameter of the 'BlueMoss' instance.

    Example. Let's assume we have matched the following html-tags against our xpath: [tag_1, tag_2, tag_3]
        1) Range(1) would filter for all tags from index 1 and onwards: [tag_2, tag_3],
           i.e. Range(x) is equivalent to the list-filtering-notation [x:]
        2) Range(0, 2) would filter for the tags at index 0 and 1: [tag_1, tag_2],
           i.e. Range(x, y) is equivalent to the list-filtering-notation [x:y].
        3) Range(1, reverse=True) would filter for the same tags as in 3.1,
           but would return them in reverse order: [tag_3, tag_2],
           i.e. Range(x, reverse=True) is equivalent to the list-filtering-notation [x:][::-1].
        4) Range(1, 10, reverse=True) would yield [tag_3, tag_2], equivalent to the notation [x:y][::-1]
    """

    start_idx: int
    end_idx: int | None = None
    reverse: bool = field(default=False)

    def __post_init__(self) -> None:
        """If end_idx is not provided, assert that start_idx is not equal to end_idx."""
        if isinstance(self.end_idx, int):
            if self.end_idx == self.start_idx:
                raise EqualIndicesException(self.start_idx)

    def filter(self, li: list[HtmlElement]) -> list[HtmlElement]:
        """Method to filter a list of elements against the configuration of this Range instance."""
        if self.end_idx is None:
            res = li[self.start_idx :]
        else:
            res = li[self.start_idx : self.end_idx]
        return res[::-1] if self.reverse else res


class EqualIndicesException(Exception):
    def __init__(self, index: int):
        message: str = (
            f'\n\nYou have defined a Range instance with start_idx={index} and end_idx={index}.'
            f"\nMake sure to provide unequal values for your 'start_idx' and 'end_idx' parameters."
            f'\n\nPro Tips:'
            f'\n1) You can define a Range instance with just one value, '
            f'e.g. Range(2) which will filter all matched html-tags from index 2 onwards.'
            f'\n2) An instance like Range(0, 2) will filter for the first two matched html-tags.'
        )
        super().__init__(message)


__all__ = ['Range', 'EqualIndicesException']



import pytest
from src.bluemoss import Root, Node, PartialKeysException


def test_partial_keys_exception():
    Root(nodes=[])
    Root(nodes=[Node(key='some_key')])
    with pytest.raises(PartialKeysException):
        Root(nodes=[Node(), Node(key='some_key')])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
from typing import List
from pathlib import Path
from setuptools import setup, find_packages


def requirements(name: str) -> List[str]:
    root = Path(__file__).parent / 'requirements'
    return root.joinpath(name).read_text().splitlines()


with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()


with open('src/bluemoss/__init__.py', encoding='utf-8') as f:
    match = re.search(
        r"^__version__\s=\s'([\d\.]+(?:-[a-zA-Z]+(?:\.\d+)?)?)'$",
        f.read(),
        re.MULTILINE,
    )
    if not match:
        raise RuntimeError('version is not set')

    version = match.group(1)

if not version:
    raise RuntimeError('version is not set')


setup(
    name='bluemoss',
    version=version,
    author='Luca Tabone',
    author_email='luca@tabone.io',
    maintainer='Luca Tabone',
    license='APACHE',
    url='https://github.com/LucaTabone/bluemoss',
    description='bluemoss enables you to easily scrape websites.',
    install_requires=requirements('base.txt'),
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(
        where='src',
        include=['bluemoss', 'bluemoss.*'],
    ),
    package_dir={'': 'src'},
    python_requires='>=3.9.0',
    package_data={'': []},
    include_package_data=True,
    zip_safe=False,
    extras_require={},
    entry_points={
        'console_scripts': [],
        'bluemoss': [],
    },
    project_urls={
        'Documentation': 'https://github.com/LucaTabone/bluemoss/blob/main/README.md',
        'Source': 'https://github.com/LucaTabone/bluemoss',
        'Tracker': 'https://github.com/LucaTabone/bluemoss/issues',
    },
    keywords=[
        'xpath',
        'scrape',
        'scraping',
        'bluemoss',
        'webscraping',
        'web-scraping',
        'recipe-scraping',
        'template-scraping',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Typing :: Typed',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Markup :: HTML',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Utilities',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Internet :: WWW/HTTP :: Browsers',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
    ],
)
